import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.imageio.*;
import javax.vecmath.*;

public class Omgifos
{
	BufferedImage image;
	Matrix4f thisFrameCam = new Matrix4f();
	Matrix4f lastFrameCam = new Matrix4f();
	Matrix4f thisFrameCamInv = new Matrix4f();
	Matrix4f lastFrameCamInv = new Matrix4f();
	float thisFrameAngle;
	float lastFrameAngle;
	Config config = new Config();
	Star[][][][] stars;
	public static void main (String[] args)
	{
		Omgifos o = new Omgifos(args);
	}
	
	public Omgifos(String[] args)
	{
		if (args.length == 3 && args[0].equals("-d"))
		{
			readConfig("omgifos.cfg",config);
			thisFrameAngle = readCamera("omgifos.cam",1,thisFrameCam);
			lastFrameAngle = readCamera("omgifos.cam",0,lastFrameCam);
			config.width = (new Integer(args[1])).intValue();
			config.height = (new Integer(args[2])).intValue();
			generateStars();
			Display display = new Display();
			display.setSize(config.width,config.height);
			display.setResizable(false);
			display.setVisible(true);
		}
		else if (args.length == 2)
		{
			readConfig("omgifos.cfg",config);
			thisFrameAngle = readCamera("omgifos.cam",1,thisFrameCam);
			lastFrameAngle = readCamera("omgifos.cam",0,lastFrameCam);
			config.width = (new Integer(args[0])).intValue();
			config.height = (new Integer(args[1])).intValue();
			generateStars();
			BufferedImage image = renderImage(config.width,config.height);
			try
			{
				ImageIO.write(image,"PNG",new File("omgifos.png"));
			}
			catch(Exception e)
			{
				System.err.println(e);
			}
		}
		else
		{
			System.out.println("usage: java -jar omgifos.jar [-d] <width> <height>");
			System.out.println("  -d\trender to display instead of file");
		}
		
		/*
		else
		{
			BufferedImage image = renderImage(640,480);
			try
			{
				ImageIO.write(image,"PNG",new File("test.png"));
			}
			catch(Exception e)
			{
				System.err.println(e);
			}
		}
		*/	
	}
	
	private BufferedImage renderImage(int width, int height)
	{
		BufferedImage image = new BufferedImage(width,height,BufferedImage.TYPE_3BYTE_BGR);
		Graphics g = image.getGraphics();
		render(g);
		return image;
	}
		
	private void render(Graphics g)
	{
		float fadeOut;
		float tsx;
		float tsy;
		float lsx;
		float lsy;
		float tfl = 1/(float)Math.tan(thisFrameAngle/360*Math.PI);
		float lfl = 1/(float)Math.tan(lastFrameAngle/360*Math.PI);
		//focalLength = 1;
		if (config.sectors == 1) fadeOut = 0; else fadeOut = 2/config.sectorSize/(config.sectors - 1);
		thisFrameCamInv.invert(thisFrameCam);
		lastFrameCamInv.invert(lastFrameCam);
		Point3f thisPos = new Point3f();
		Point3f lastPos = new Point3f();
		Graphics2D g2 = (Graphics2D)g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL,RenderingHints.VALUE_STROKE_PURE);
		g2.setRenderingHint(RenderingHints.KEY_RENDERING,RenderingHints.VALUE_RENDER_QUALITY);
		g2.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING,RenderingHints.VALUE_COLOR_RENDER_QUALITY);
		g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION,RenderingHints.VALUE_INTERPOLATION_BICUBIC);
		g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
		//g2.setBackground(new Color(0,0,0));
		//g2.clearRect(0,0,config.width - 1,config.height - 1);
		g2.setStroke(new BasicStroke(config.starSize,BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND));
		g2.setColor(new Color(255,0,0));
		for (int x = 0; x < config.sectors; x++)
		{
			for (int y = 0; y < config.sectors; y++)
			{
				for (int z = 0; z < config.sectors; z++)
				{
					for (int s = 0; s < config.starsPerSector; s++)
					{
						thisPos.set(stars[x][y][z][s].position);
						thisFrameCamInv.transform(thisPos);
						float distance = thisPos.distance(new Point3f(0,0,0));
						thisPos.x = thisPos.x*tfl/thisPos.z;
						thisPos.y = thisPos.y*tfl/thisPos.z; 
						tsx = config.width/2 + thisPos.x*config.width;
						tsy = config.height/2 - thisPos.y*config.height*config.aspectRatio;
						lastPos.set(stars[x][y][z][s].position);
						//lastPos.sub(new Vector3f(0,0,500));
						lastFrameCamInv.transform(lastPos);
						lastPos.x = lastPos.x*lfl/lastPos.z;
						lastPos.y = lastPos.y*lfl/lastPos.z;
						lsx = config.width/2 + lastPos.x*config.width;
						lsy = config.height/2 - lastPos.y*config.height*config.aspectRatio;
						float screenDistance = (float)Math.sqrt((lsx-tsx)*(lsx-tsx)+(lsy-tsy)*(lsy-tsy));
						if (screenDistance < 1) screenDistance = 1;
						float motionFade = 1.6f/screenDistance + 0.3f;
						float alpha = (1 - distance*fadeOut)*motionFade;
						if (alpha < 0) alpha = 0;
						if (alpha > 1) alpha = 1;
						Color3f col = stars[x][y][z][s].color;
						//System.out.println(col.x+","+col.y+","+col.z);
						//System.out.println(thisPos.x + "," + thisPos.y + "," + thisPos.z);
						g2.setColor(new Color(col.x,col.y,col.z,alpha));
						if (thisPos.z > tfl*100 && lastPos.z > lfl*100) g2.draw(new Line2D.Float(tsx,tsy,lsx,lsy));
					}
				}
			}
		}
						
	}
	
	private boolean readConfig(String configFile, Config config)
	{
		BufferedReader file;
		String line;
		int lineNr = 0;
		try
		{
			file = new BufferedReader(new FileReader(configFile));
			while ((line = file.readLine()) != null)
			{
				lineNr++;
				String[] part = line.split(":");
				if (part.length == 2)
				{
					String attribute = part[0];
					String value = part[1];
					try
					{
						if(attribute.startsWith("Width")) config.width = (new Integer(value)).intValue();
						if(attribute.startsWith("Height")) config.height = (new Integer(value)).intValue();
						if(attribute.startsWith("AspectRatio")) { 
							String[] ar = value.split("/");
							config.aspectRatio = (new Float(ar[0])).floatValue()/(new Float(ar[1])).floatValue();
						}
						if(attribute.startsWith("StarSize")) config.starSize = (new Float(value)).floatValue();
						if(attribute.startsWith("MotionBlur")) config.motionBlur = (value.equals("on"));
						if(attribute.startsWith("MotionDarkening")) config.motionDarkening = (new Float(value)).floatValue();
						
						if(attribute.startsWith("Sectors")) config.sectors = (new Integer(value)).intValue();
						if(attribute.startsWith("SectorSize")) config.sectorSize = (new Float(value)).floatValue();
						if(attribute.startsWith("StarsPerSector")) config.starsPerSector = (new Integer(value)).intValue();
						if(attribute.startsWith("Repeat")) config.repeat = (new Integer(value)).intValue();
						if(attribute.startsWith("RandomSeed")) config.randomSeed = (new Integer(value)).intValue();
					}
					catch(Exception e)
					{
						System.err.println(e);
						error("Error in " + configFile + " line " + lineNr + "\n" + line);
					}
				}
			}
		}
		catch (Exception e)
		{
			error(e.toString());
		}
		return false;
	}
	
	private float readCamera(String cameraFile, int frame, Matrix4f cam)
	{
		BufferedReader file;
		String line;
		int lineNr = 0;
		int f = -1;
		try
		{
			file = new BufferedReader(new FileReader(cameraFile));
			while ((line = file.readLine()) != null)
			{
				lineNr++;
				String[] part = line.split(":");
				try
				{
					f = (new Integer(part[0])).intValue();
				}
				catch(NumberFormatException e)
				{
					error("Error in config line " + lineNr + "\n" + line);
				}
				if (f == frame)
				{
					String value = part[1];
					value = value.replaceAll("<","");
					value = value.replaceAll(">","");
					if(!toMatrix4f(value,cam)) error("Error in config line " + lineNr + "\n" + line);
					float angle = (new Float(part[2])).floatValue();
					config.aspectRatio = (new Float(part[3])).floatValue();
					file.close();
					return angle;
				}
			}
			error("frame# " + frame + " not found");
			file.close();
		}
		catch (Exception e)
		{
			error(e.toString());
		}
		return -1;
	}
	
	private void error(String e)
	{
		System.err.println(e);
		System.exit(1);
	}
	
	private boolean toTuple3f(String s,Tuple3f t)
	{
		String[] comp = s.split(",");
		if (comp.length == 3)
		{
			float x = (new Float(comp[0])).floatValue();
			float y = (new Float(comp[1])).floatValue();
			float z = (new Float(comp[2])).floatValue();
			t.set(x,y,z);
			return true;
		}
		else return false;
	}
	
	private boolean toMatrix4f(String s,Matrix4f m)
	{
		String[] comp = s.split(",");
		if (comp.length == 12)
		{
			float[] e = new float[16];
			for (int i = 0; i < 12; i++) e[i] = (new Float(comp[i])).floatValue();
			e[12] = 0;
			e[13] = 0;
			e[14] = 0;
			e[15] = 1;
			m.set(e);
			return true;
		}
		else return false;
	}
	
	private Point3f getCamLocation()
	{
		Point3f camPos = new Point3f();
		thisFrameCam.transform(camPos);
		return camPos;
	}
		
	private void generateStars()
	{
		Random rnd;
		float m;
		float o;
		int xSector;
		int ySector;
		int zSector;
		Point3f camPos;
		stars = new Star[config.sectors][config.sectors][config.sectors][config.starsPerSector];
		camPos = getCamLocation();
		m = config.sectors/2;
		o = config.sectors%2/2*config.sectorSize;
		if (camPos.x < 0) xSector = (int)((camPos.x + o)/config.sectorSize) - 1; else xSector = (int)((camPos.x + o)/config.sectorSize);
		if (camPos.y < 0) ySector = (int)((camPos.y + o)/config.sectorSize) - 1; else ySector = (int)((camPos.y + o)/config.sectorSize);
		if (camPos.z < 0) zSector = (int)((camPos.z + o)/config.sectorSize) - 1; else zSector = (int)((camPos.z + o)/config.sectorSize);
		for (int x = 0; x < config.sectors; x++)
		{
			for (int y = 0; y < config.sectors; y++)
			{
				for (int z = 0; z < config.sectors; z++)
				{
					rnd = new Random(config.randomSeed + 100*((xSector + x)%config.repeat) + 100*config.repeat*((ySector + y)%config.repeat) + 100*config.repeat*config.repeat*((zSector + z)%config.repeat));
					for (int s = 0; s < config.starsPerSector; s++)
					{
						stars[x][y][z][s] = new Star();
						float ch = 2*(rnd.nextFloat()%1-0.5f);
						float cs = Math.abs(rnd.nextFloat()%1);
						float cv = Math.abs(rnd.nextFloat()%0.5f) + 0.5f;
						float rg;
						float b;
						if (ch < 0)
						{
							rg = (-ch*(1 - cs) + cs)*cv;
							b = cv;
						}
						else
						{
							b = (ch*(1 - cs) + cs)*cv;
							rg = cv;
						}
						stars[x][y][z][s].color = new Color3f(rg,rg,b);
						Vector3f pos = new Vector3f(Math.abs(rnd.nextFloat()%1),Math.abs(rnd.nextFloat()%1),Math.abs(rnd.nextFloat()%1));
						pos.add(new Vector3f(xSector + x - m, ySector + y - m, zSector + z - m));
						pos.scale(config.sectorSize);
						stars[x][y][z][s].position = pos;
					}
				}
			}
		}		
	}
	
	private String getFrameNr(int f) {
		String nr = "" + f;
		if (f < 10000) nr = "0" + nr;
		if (f < 1000) nr = "0" + nr;
		if (f < 100) nr = "0" + nr;
		if (f < 10) nr = "0" + nr;
		return nr;
	}
	
	class Display
	extends Frame
	{
		public Display()
		{
			super("OMGIFOS");
			GraphicsPanel gp = new GraphicsPanel();
			add(gp);
			addWindowListener(new WindowClosingAdapter());
		}
	}
	
	class WindowClosingAdapter
	extends WindowAdapter
	{
		public void windowClosing(WindowEvent event)
		{
			event.getWindow().setVisible(false);
			event.getWindow().dispose();
			System.exit(0);
		}
	}
	
	class GraphicsPanel
	extends Panel
	{
		public void paint(Graphics g)
		{
			super.paint(g);
			Graphics2D g2 = (Graphics2D)g;
			g2.setBackground(new Color(0,0,0));
			g2.clearRect(0,0,config.width - 1,config.height - 1);
			render(g);
		}
	}
	
	class Config
	{
		int width;
		int height;
		float aspectRatio;
		float starSize;
		boolean motionBlur;
		float motionDarkening;
		int mode;
		int sectors;
		float sectorSize;
		int starsPerSector;
		int repeat;
		int randomSeed;
	}
	
	class Star
	{
		Vector3f position;
		Color3f color;
	}
}
	
	