/*
*	The Eye of Japetus had blinked, as if to remove an
*	irritating speck of dust. David Bowman had time for
*	just one broken sentence, which the waiting men in
*	Mission Control, nine hundred million miles away and
*	ninety minutes in the future, were never to forget:
*	The thing's hollow - it goes on for ever - and -
*	oh my God - IT'S FULL OF STARS!
*
*	===================
*	OMGIFOS version 1.0
*	===================
*
*	Copyright (c) 2003 Sascha Ledinsky
*/

#include "transforms.inc"

#macro omgifosSetCamera(_cameraLocation,_cameraLookAt,_cameraAngle,_cameraTransform,_cameraAspectRatio)
	#declare cameraLocation = _cameraLocation;
	#declare cameraLookAt = _cameraLookAt;
	#declare cameraAngle = _cameraAngle;
	#declare cameraAspectRatio = _cameraAspectRatio;
	#declare cameraTransformation = transform {
		Reorient_Trans(z,(cameraLookAt - cameraLocation))
		transform _cameraTransform
		translate cameraLocation
	}
	camera {
		up y
		right x*cameraAspectRatio
		transform cameraTransformation
		angle cameraAngle
	}
#end

#macro omgifosSetMotion(_motionLocation,_motionLookAt,_motionAngle,_motionTransform)
	#declare motionLocation = _motionLocation;
	#declare motionLookAt = _motionLookAt;
	#declare motionAngle = _motionAngle;
	#declare motionTransformation = transform
	{
		transform _motionTransform
		Reorient_Trans(z,(motionLookAt - motionLocation))
		translate motionLocation
	}
#end

// matrix2string(float[12])
// returns: string
#macro matrix2string(m)
	concat("<",str(m[0],3,3),",",str(m[1],3,3),",",str(m[2],3,3),",",
		str(m[3],3,3),",",str(m[4],3,3),",",str(m[5],3,3),",",
		str(m[6],3,3),",",str(m[7],3,3),",",str(m[8],3,3),",",
		str(m[9],3,3),",",str(m[10],3,3),",",str(m[11],3,3),">")
#end

// getMatrix(transform tr)
// returns: string
#macro getMatrix(tr)
	#local cam = array[12];
	#local vector = vtransform(<0,0,0>,tr);
	#local cam[3] = vector.x;
	#local cam[7] = vector.y;
	#local cam[11] = vector.z;
	#local translation = vector;
	#local vector = vtransform(x,tr)-translation;
	#local cam[0] = vector.x;
	#local cam[4] = vector.y;
	#local cam[8] = vector.z;
	#local vector = vtransform(y,tr)-translation;
	#local cam[1] = vector.x;
	#local cam[5] = vector.y;
	#local cam[9] = vector.z;
	#local vector = vtransform(z,tr)-translation;
	#local cam[2] = vector.x;
	#local cam[6] = vector.y;
	#local cam[10] = vector.z;
	matrix2string(cam)
#end

// writeCamera()
#macro writeCamera()
	#fopen camDefFile "omgifos.cam" write
	#write (camDefFile,concat("0:",getMatrix(cameraTransformation),":",str(cameraAngle,3,3),":",str(cameraAspectRatio,5,5),"\n"))
	#write (camDefFile,concat("1:",getMatrix(motionTransformation),":",str(motionAngle,3,3),":",str(cameraAspectRatio,5,5),"\n"))
	#fclose camDefFile
#end

#macro setScreen(omgifosDistance)
	plane {
		-z,-omgifosDistance
		pigment {
			image_map {
				png "omgifos.png"
				once
			}
			translate <-0.5,-0.5,0>
			//rotate <0,180,0>
			scale tan(cameraAngle/360*pi)*omgifosDistance
			scale <2,2/cameraAspectRatio,1>
		}
		finish {
			ambient 1
			diffuse 0
			specular 0
			phong 0
			reflection 0
		}
		transform {
			cameraTransformation
			//inverse
		}
	}
#end